"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.decodeToken = exports.hasTokenExpired = exports.isRefreshTokenExpiring = exports.isAccessTokenExpiring = void 0;
const jwt_decode_1 = __importDefault(require("jwt-decode"));
/**
 * Checks if the Access Token is expired or is expiring in 1 hour.  A default Access Token
 * lasts 12 hours. If the Access Token expires, the Refresh Token is used to fetch a new
 * Access Token. In the case that the Refresh Token is expired, 1 hour is enough to let
 * most jobs finish.
 * @param {string} token- token string that will be evaluated
 * @param {number} timeToLiveSeconds - the amount of time that the token has before it expires, defaults to 3600
 * @returns {boolean} a value indicating whether the token is about to expire
 */
function isAccessTokenExpiring(token, timeToLiveSeconds = 3600) {
    if (!token) {
        return true;
    }
    return isTokenExpiring(token, timeToLiveSeconds);
}
exports.isAccessTokenExpiring = isAccessTokenExpiring;
/**
 * Checks if the Refresh Token is expired or expiring in 30 secs. A default Refresh Token
 * lasts 30 days.  Once the Refresh Token expires, the user must re-authenticate (provide
 * credentials in a browser to obtain an authorisation code). 30 seconds is enough time
 * to make a request for a final Access Token.
 * @param {string} token- token string that will be evaluated
 * @param {number} timeToLiveSeconds - the amount of time that the token has before it expires, defaults to 30
 * @returns {boolean} a value indicating whether the token is about to expire
 */
function isRefreshTokenExpiring(token, timeToLiveSeconds = 30) {
    if (!token) {
        return true;
    }
    return isTokenExpiring(token, timeToLiveSeconds);
}
exports.isRefreshTokenExpiring = isRefreshTokenExpiring;
/**
 * Checks if the given token has expired.
 * @param {string} token- token string that will be evaluated
 * @returns {boolean} a value indicating whether the token has expired
 */
function hasTokenExpired(token) {
    if (!token) {
        return true;
    }
    return isTokenExpiring(token, 0);
}
exports.hasTokenExpired = hasTokenExpired;
function isTokenExpiring(token, timeToLiveSeconds) {
    const payload = (0, jwt_decode_1.default)(token);
    const timeToLive = payload.exp - new Date().valueOf() / 1000;
    return timeToLive <= timeToLiveSeconds;
}
function decodeToken(token) {
    let payload;
    try {
        payload = (0, jwt_decode_1.default)(token);
    }
    catch (err) {
        throw new Error('Invalid token supplied.');
    }
    return payload;
}
exports.decodeToken = decodeToken;
//# sourceMappingURL=auth.js.map