"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.updateCsv = exports.createCsv = exports.readCsv = void 0;
const file_1 = require("./file");
const format_1 = require("@fast-csv/format");
/**
 * reads a CSV file and returns parsed data
 * @param csvFilePath
 * @returns string[][]
 */
const readCsv = (csvFilePath) => __awaiter(void 0, void 0, void 0, function* () {
    const csvContent = yield (0, file_1.readFile)(csvFilePath);
    return csvContent
        .split('\n')
        .filter((row) => row.length)
        .map((data) => data.split(','));
});
exports.readCsv = readCsv;
/**
 * creates a csv file at given path with provided data
 * @param csvFilePath location where to create file
 * @param csvData data which needs to be populated in file
 * @param headers an array of column names
 */
const createCsv = (csvFilePath, csvData, headers) => __awaiter(void 0, void 0, void 0, function* () {
    const output = yield (0, format_1.writeToString)(csvData, { headers });
    yield (0, file_1.createFile)(csvFilePath, output);
});
exports.createCsv = createCsv;
/**
 * append new csv record if file exists else creates a new file
 * @param csvFilePath location where to create file
 * @param newRecord new record which needs to be appended
 * @param columns an array of header for file, i.e column names
 * @param prependId name of id column
 */
const updateCsv = (csvFilePath, newRecord, columns, prependId) => __awaiter(void 0, void 0, void 0, function* () {
    const csvData = yield validateInput(csvFilePath, newRecord, columns, prependId);
    if (prependId) {
        const newId = csvData.length === 0 ? 1 : csvData.length;
        const idIndexInColumns = columns.findIndex((col) => col === prependId);
        if (idIndexInColumns > -1) {
            newRecord.splice(idIndexInColumns, 0, newId);
        }
        else {
            columns.splice(0, 0, prependId);
            newRecord.splice(0, 0, newId);
        }
    }
    csvData.push(newRecord);
    yield (0, exports.createCsv)(csvFilePath, csvData, columns);
});
exports.updateCsv = updateCsv;
const validateInput = (csvFilePath, newRecord, columnsProvided, prependId) => __awaiter(void 0, void 0, void 0, function* () {
    if (newRecord.length !== columnsProvided.length) {
        if (newRecord.length > columnsProvided.length) {
            throw new Error('a record can not have more fields than provided columns');
        }
        // provided columns can/cannot have 'id'
        if (!(prependId && newRecord.length + 1 === columnsProvided.length)) {
            throw new Error('a record can not have less fields than provided columns');
        }
    }
    else if (prependId && columnsProvided.includes(prependId)) {
        throw new Error('a record can not have more fields than provided columns');
    }
    const csvData = yield (0, exports.readCsv)(csvFilePath).catch((_) => []);
    const columnsInFile = csvData[0];
    if (columnsInFile) {
        if (columnsProvided.length > columnsInFile.length) {
            throw new Error('number of provided columns are greater than number of existing columns');
        }
        if (columnsProvided.length < columnsInFile.length) {
            // provided columns can/cannot have 'id'
            if (!(prependId && columnsProvided.length + 1 === columnsInFile.length))
                throw new Error('number of provided columns are less than number of existing columns');
            const idIndexInColumns = columnsInFile.findIndex((col) => col === prependId);
            const columnsWithId = [...columnsProvided];
            if (idIndexInColumns > -1) {
                columnsWithId.splice(idIndexInColumns, 0, prependId);
                if (JSON.stringify(columnsProvided) !== JSON.stringify(columnsWithId)) {
                    throw new Error('provided columns does not match existing columns');
                }
            }
            else {
                throw new Error('provided columns does not match existing columns');
            }
        }
        else {
            if (JSON.stringify(columnsProvided) !== JSON.stringify(columnsInFile)) {
                throw new Error('provided columns does not match existing columns');
            }
            if (prependId && !columnsProvided.includes(prependId)) {
                throw new Error('can not add new id to existing data');
            }
        }
    }
    return csvData;
});
//# sourceMappingURL=csvFile.js.map