"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.extractHashArray = exports.compareHashes = exports.getHash = void 0;
const fs_extra_1 = __importDefault(require("fs-extra"));
const path_1 = __importDefault(require("path"));
const crypto_1 = require("crypto");
const file_1 = require("../file");
/**
 * Hashes each file in each directory, and then hashes the hashes to create a hash for each directory also.
 *
 * Whilst files are hashed in their entirety, the logic for creating a folder hash is as follows:
 *
 * Sort the files and subfolders by name (case sensitive, uppercase then lower)
 * Take the first 100 hashes, concatenate and hash
 * Concatenate this hash with another 100 hashes and hash again
 * Continue until the end of the folder. This is the folder hash
 * If a folder contains other folders, start from the bottom of the tree - the folder hashes cascade upwards so you know immediately if there is a change in a sub/sub directory
 * If the folder has no content (empty) then it is ignored. No hash created.
 *
 * @param {string} folderPath - absolute folder path
 */
const getHash = (folderPath) => __awaiter(void 0, void 0, void 0, function* () {
    return yield hashFolder(folderPath);
});
exports.getHash = getHash;
/**
 * It returns a hashed folder tree that contains the local directory resources that are not synced with remote
 */
function compareHashes(localHash, remoteHashMap) {
    return hashDifference(localHash, remoteHashMap);
}
exports.compareHashes = compareHashes;
const extractHashArray = (log) => {
    if (log.includes('>>weboutBEGIN<<')) {
        try {
            const webout = log
                .split(/>>weboutBEGIN<<[\n|\r\n]/)[1]
                .split(/>>weboutEND<<[\n|\r\n]/)[0];
            const weboutWithoutLF = webout.replace(/\n|\r\n/g, '');
            const jsonWebout = JSON.parse(weboutWithoutLF);
            return jsonWebout.hashes;
        }
        catch (err) {
            throw new Error(`An error occurred while extracting hashes array from webout: ${err.message}`);
        }
    }
};
exports.extractHashArray = extractHashArray;
const hashDifference = (localHash, remoteHashMap, hashedDiff = {
    hash: localHash.hash,
    absolutePath: localHash.absolutePath,
    relativePath: localHash.relativePath,
    isFile: false,
    members: []
}) => {
    for (const member of localHash.members) {
        if (remoteHashMap[member.relativePath] !== member.hash) {
            if (member.isFile) {
                hashedDiff.members.push(member);
            }
            else {
                const diff = hashDifference(member, remoteHashMap, {
                    hash: member.hash,
                    absolutePath: member.absolutePath,
                    relativePath: member.relativePath,
                    isFile: false,
                    members: []
                });
                hashedDiff.members.push(diff);
            }
        }
    }
    return hashedDiff;
};
const hashFile = (filePath) => __awaiter(void 0, void 0, void 0, function* () {
    const fileContent = fs_extra_1.default.readFileSync(filePath);
    if (fileContent.length) {
        const hash = (0, crypto_1.createHash)('md5');
        hash.update(fileContent);
        return hash.digest('hex').toUpperCase();
    }
});
const hashFolder = (folderPath, pathRelativeTo = folderPath, hashedFolder = {
    hash: '',
    absolutePath: folderPath,
    relativePath: (0, file_1.getRelativePath)(pathRelativeTo, folderPath),
    isFile: false,
    members: []
}) => __awaiter(void 0, void 0, void 0, function* () {
    const filesAndFolders = yield (0, file_1.listFilesAndSubFoldersInFolder)(folderPath, false);
    filesAndFolders.sort();
    let concatenatedHash = '';
    const chunkSize = 100;
    for (let i = 0; i < filesAndFolders.length; i += chunkSize) {
        // Take the first 100 hashes, concatenate and hash
        const resources = filesAndFolders.slice(i, i + chunkSize);
        for (const resource of resources) {
            const resourcePath = path_1.default.join(folderPath, resource);
            if (yield (0, file_1.isFolder)(resourcePath)) {
                const hashedSubFolder = yield hashFolder(resourcePath, pathRelativeTo);
                hashedFolder.members.push(hashedSubFolder);
                concatenatedHash += hashedSubFolder.hash;
            }
            else {
                const hash = yield hashFile(resourcePath);
                if (!hash)
                    continue;
                hashedFolder.members.push({
                    hash,
                    absolutePath: resourcePath,
                    relativePath: (0, file_1.getRelativePath)(pathRelativeTo, resourcePath),
                    isFile: true
                });
                concatenatedHash += hash;
            }
        }
        if (concatenatedHash !== '') {
            const hash = (0, crypto_1.createHash)('md5');
            hash.update(concatenatedHash);
            concatenatedHash = hash.digest('hex').toUpperCase();
        }
    }
    hashedFolder.hash = concatenatedHash;
    return hashedFolder;
});
//# sourceMappingURL=hash.js.map