"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Target = void 0;
const config_1 = require("./config");
const serverType_1 = require("./serverType");
const targetValidators_1 = require("./targetValidators");
class Target {
    constructor(json, config = {}) {
        this._serverType = serverType_1.ServerType.SasViya;
        this.getConfig = (key, json) => {
            const config = this._config[key] || {};
            return Object.assign(Object.assign({}, config), json[key]);
        };
        try {
            if (!json) {
                throw new Error('Invalid target: Input JSON is null or undefined.');
            }
            this._config = config;
            this._name = (0, targetValidators_1.validateTargetName)(json.name);
            this._serverUrl = (0, targetValidators_1.validateServerUrl)(json.serverUrl);
            this._serverType = (0, targetValidators_1.validateServerType)(json.serverType);
            this._httpsAgentOptions = (0, targetValidators_1.validateHttpsAgentOptions)(json.httpsAgentOptions);
            this._appLoc = (0, targetValidators_1.validateAppLoc)(json.appLoc);
            this._contextName = (0, targetValidators_1.validateContextName)(json.contextName, this._serverType);
            this._serverName = (0, targetValidators_1.validateServerName)(json.serverName, this._serverType);
            this._syncFolder = (0, targetValidators_1.validateSyncFolder)(json.syncFolder);
            this._repositoryName = (0, targetValidators_1.validateRepositoryName)(json.repositoryName, this._serverType);
            if (json.docConfig) {
                this._docConfig = (0, targetValidators_1.validateDocConfig)(this.getConfig(config_1.ConfigTypes.Doc, json));
            }
            if (json.authConfig) {
                this._authConfig = (0, targetValidators_1.validateAuthConfig)(this.getConfig(config_1.ConfigTypes.Auth, json));
            }
            if (json.authConfigSas9) {
                this._authConfigSas9 = (0, targetValidators_1.validateAuthConfigSas9)(json.authConfigSas9);
            }
            if (json.buildConfig) {
                this._buildConfig = (0, targetValidators_1.validateBuildConfig)(this.getConfig(config_1.ConfigTypes.Build, json), this._name);
            }
            if (json.deployConfig) {
                this._deployConfig = (0, targetValidators_1.validateDeployConfig)(this.getConfig(config_1.ConfigTypes.Deploy, json));
            }
            else {
                this._deployConfig = (0, targetValidators_1.validateDeployConfig)({
                    deployServicePack: true,
                    deployScripts: []
                });
            }
            if (json.serviceConfig) {
                this._serviceConfig = (0, targetValidators_1.validateServiceConfig)(this.getConfig(config_1.ConfigTypes.Service, json));
            }
            if (json.jobConfig) {
                this._jobConfig = (0, targetValidators_1.validateJobConfig)(this.getConfig(config_1.ConfigTypes.Job, json));
            }
            if (json.streamConfig) {
                this._streamConfig = (0, targetValidators_1.validateStreamConfig)(this.getConfig(config_1.ConfigTypes.Stream, json));
            }
            if (json.testConfig) {
                this._testConfig = (0, targetValidators_1.validateTestConfig)(this.getConfig(config_1.ConfigTypes.Test, json));
            }
            if (json.macroFolders && json.macroFolders.length) {
                this._macroFolders = json.macroFolders;
            }
            if (json.programFolders && json.programFolders.length) {
                this._programFolders = json.programFolders;
            }
            if (json.binaryFolders && json.binaryFolders.length) {
                this._binaryFolders = json.binaryFolders;
            }
            if (json.syncDirectories && json.syncDirectories.length) {
                this._syncDirectories = (0, targetValidators_1.validateSyncDirectories)(json.syncDirectories);
            }
            if (json.sasjsBuildFolder) {
                this._sasjsBuildFolder = (0, targetValidators_1.validateSasjsBuildFolder)(json.sasjsBuildFolder);
            }
            if (json.sasjsResultsFolder) {
                this._sasjsResultsFolder = (0, targetValidators_1.validateSasjsResultsFolder)(json.sasjsResultsFolder);
            }
        }
        catch (e) {
            throw new Error(`Error parsing target: ${e.message}`);
        }
    }
    get name() {
        return this._name;
    }
    get serverUrl() {
        return this._serverUrl;
    }
    get serverType() {
        return this._serverType;
    }
    get httpsAgentOptions() {
        return this._httpsAgentOptions;
    }
    get appLoc() {
        return this._appLoc;
    }
    get docConfig() {
        return this._docConfig;
    }
    get authConfig() {
        return this._authConfig;
    }
    get authConfigSas9() {
        return this._authConfigSas9;
    }
    get buildConfig() {
        return this._buildConfig;
    }
    get deployConfig() {
        return this._deployConfig;
    }
    get serviceConfig() {
        return this._serviceConfig;
    }
    get jobConfig() {
        return this._jobConfig;
    }
    get streamConfig() {
        return this._streamConfig;
    }
    get macroFolders() {
        return this._macroFolders;
    }
    get programFolders() {
        return this._programFolders;
    }
    get binaryFolders() {
        return this._binaryFolders;
    }
    get contextName() {
        return this._contextName;
    }
    get serverName() {
        return this._serverName;
    }
    get repositoryName() {
        return this._repositoryName;
    }
    get testConfig() {
        return this._testConfig;
    }
    get syncFolder() {
        return this._syncFolder;
    }
    get syncDirectories() {
        return this._syncDirectories;
    }
    get sasjsBuildFolder() {
        return this._sasjsBuildFolder;
    }
    get sasjsResultsFolder() {
        return this._sasjsResultsFolder;
    }
    toJson(withDefaults = true) {
        var _a, _b, _c;
        const json = {
            name: this.name,
            serverUrl: this.serverUrl,
            serverType: this.serverType,
            httpsAgentOptions: this.httpsAgentOptions,
            appLoc: this.appLoc,
            docConfig: this.docConfig,
            deployConfig: this.deployConfig
        };
        if ((_a = this.macroFolders) === null || _a === void 0 ? void 0 : _a.length)
            json.macroFolders = this.macroFolders;
        if ((_b = this.programFolders) === null || _b === void 0 ? void 0 : _b.length)
            json.programFolders = this.programFolders;
        if ((_c = this.binaryFolders) === null || _c === void 0 ? void 0 : _c.length)
            json.binaryFolders = this.binaryFolders;
        if (this.authConfig) {
            json.authConfig = this.authConfig;
        }
        if (this.authConfigSas9) {
            json.authConfigSas9 = this.authConfigSas9;
        }
        if (this.sasjsBuildFolder) {
            json.sasjsBuildFolder = this.sasjsBuildFolder;
        }
        else if (withDefaults) {
            json.sasjsBuildFolder = targetValidators_1.DEFAULT_SASJS_BUILD_FOLDER;
        }
        if (this.sasjsResultsFolder) {
            json.sasjsResultsFolder = this.sasjsResultsFolder;
        }
        else if (withDefaults) {
            json.sasjsResultsFolder = targetValidators_1.DEFAULT_SASJS_RESULTS_FOLDER;
        }
        if (this.syncDirectories) {
            json.syncDirectories = this.syncDirectories;
        }
        else if (withDefaults) {
            json.syncDirectories = [];
        }
        if (this.buildConfig) {
            json.buildConfig = this.buildConfig;
        }
        else if (withDefaults)
            json.buildConfig = {
                initProgram: '',
                termProgram: '',
                buildOutputFileName: `${this.name}.sas`,
                macroVars: {}
            };
        if (this.jobConfig) {
            json.jobConfig = this.jobConfig;
        }
        else if (withDefaults)
            json.jobConfig = {
                jobFolders: [],
                initProgram: '',
                termProgram: '',
                macroVars: {}
            };
        if (this.serviceConfig) {
            json.serviceConfig = this.serviceConfig;
        }
        else if (withDefaults)
            json.serviceConfig = {
                serviceFolders: [],
                initProgram: '',
                termProgram: '',
                macroVars: {}
            };
        if (this.streamConfig) {
            json.streamConfig = this.streamConfig;
        }
        else if (withDefaults)
            json.streamConfig = {
                streamWebFolder: '',
                streamWeb: false,
                webSourcePath: '',
                streamServiceName: '',
                assetPaths: []
            };
        if (this.testConfig) {
            json.testConfig = this.testConfig;
        }
        else if (withDefaults)
            json.testConfig = {
                initProgram: '',
                termProgram: '',
                macroVars: {},
                testSetUp: '',
                testTearDown: ''
            };
        if (this.serverType === serverType_1.ServerType.SasViya) {
            json.contextName = this.contextName;
        }
        else if (this.serverType === serverType_1.ServerType.Sas9) {
            json.serverName = this.serverName;
            json.repositoryName = this.repositoryName;
        }
        return json;
    }
}
exports.Target = Target;
//# sourceMappingURL=target.js.map